/*-
 * Copyright (c) 2013, A P Francisco
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _BINHEAP_H_
#define _BINHEAP_H_

#include <limits.h>
#include <stdlib.h>
#include <string.h>

struct binheap {
  size_t   size;
  size_t   max_size;
  size_t  *heap;
  size_t  *map;
  void    *key;
  size_t   ksz;
  int    (*cmp)(const void *,const void *);
};

#define LEFT(x)    ((((x) + 1) << 1) - 1)
#define RIGHT(x)   (((x) + 1) << 1)
#define PARENT(x)  ((((x) + 1) >> 1) - 1)

#define  BINHEAP_INIT(q, type, k, n, c) do {                              \
  (q) = malloc(sizeof(struct binheap));                                   \
  (q)->max_size = (n);                                                    \
  (q)->size = 0;                                                          \
  (q)->heap = malloc(sizeof(size_t)*(n));                                 \
  memset((q)->heap, 0xFF, sizeof(size_t)*(n));                            \
  (q)->map = malloc(sizeof(size_t)*(n));                                  \
  memset((q)->map, 0xFF, sizeof(size_t)*(n));                             \
  (q)->key = (k);                                                         \
  (q)->ksz = sizeof(type);                                                \
  (q)->cmp = (c);                                                         \
} while (/*CONSTCOND*/0)

#define  BINHEAP_FREE(q) do {                                             \
  free((q)->heap);                                                        \
  free((q)->map);                                                         \
  free(q);                                                                \
} while (/*CONSTCOND*/0)

#define  BINHEAP_EMPTY(q)  ((q)->size == 0)
#define BINHEAP_HAS(q, id)  ((q)->map[id] < (q)->size)
#define BINHEAP_TOP(q)    ((q)->heap[0])

#define BINHEAP_RESET(q) do {                                             \
  (q)->size = 0;                                                          \
  memset((q)->heap, 0xFF, sizeof(size_t)*(q)->max_size);                  \
  memset((q)->map, 0xFF, sizeof(size_t)*(q)->max_size);                   \
} while (/*CONSTCOND*/0)

#define BINHEAP_PUSH(q, id) do {                                          \
  if ((q)->map[id] >= ULONG_MAX) {                                        \
    if ((q)->size >= (q)->max_size)                                       \
      abort();                                                            \
                                                                          \
    (q)->map[id] = (q)->size;                                             \
    (q)->heap[(q)->size] = (id);                                          \
    (q)->size ++;                                                         \
  }                                                                       \
  binheap_siftup((q)->key, (q)->heap, (q)->map, (q)->size,                \
      (q)->ksz, (q)->map[id], (q)->cmp);                                  \
  binheap_siftdown((q)->key, (q)->heap, (q)->map, (q)->size,              \
      (q)->ksz, (q)->map[id], (q)->cmp);                                  \
} while (/*CONSTCOND*/0)

#define BINHEAP_UPDATE(q, id)  BINHEAP_PUSH(q, id)

#define BINHEAP_DEL(q, id)  do {                                          \
  register size_t aux = (q)->map[id];                                     \
  q->size--;                                                              \
  if (q->size != aux) {                                                   \
    binheap_exchange((q)->heap, (q)->map, (q)->size, aux);                \
    binheap_siftup((q)->key, (q)->heap, (q)->map, (q)->size,              \
        (q)->ksz, aux, (q)->cmp);                                         \
    binheap_siftdown((q)->key, (q)->heap, (q)->map, (q)->size,            \
        (q)->ksz, aux, (q)->cmp);                                         \
  }                                                                       \
  q->heap[q->size] = ULONG_MAX;                                           \
  q->map[id] = ULONG_MAX;                                                 \
} while (/*CONSTCOND*/0)

#define BINHEAP_POP(q)  do {                                              \
  register size_t aux = q->heap[0];                                       \
  q->size--;                                                              \
  binheap_exchange((q)->heap, (q)->map, (q)->size, 0);                    \
  binheap_siftdown((q)->key, (q)->heap, (q)->map, (q)->size,              \
      (q)->ksz, 0, (q)->cmp);                                             \
  q->heap[q->size] = ULONG_MAX;                                           \
  q->map[aux] = ULONG_MAX;                                                \
} while (/*CONSTCOND*/0)

#define BINHEAP_MAKE(q, n)  do {                                          \
  (q)->size = (n);                                                        \
  binheap_make((q)->key, (q)->heap, (q)->map,                             \
      (q)->size, (q)->ksz, (q)->cmp);                                     \
} while (/*CONSTCOND*/0)

void  binheap_exchange(size_t *heap, size_t *map, size_t i, size_t j);

void  binheap_make(void *key, size_t *heap, size_t *map, size_t n,
      size_t size, int(*cmp)(const void *, const void *));

void  binheap_siftdown(void *key, size_t *heap, size_t *map, size_t n,
      size_t size, size_t i, int(*cmp)(const void *, const void *));

void  binheap_siftup(void *key, size_t *heap, size_t *map, size_t n,
      size_t size, size_t i, int(*cmp)(const void *, const void *));

#endif /* _BINHEAP_H_ */
